from .classifier import ClassificationHead, ProjectionDiscr
from .vgg import *
from .resnet import *
from .patchgan import *
from .pixcnn import *
from .perceptualnet import *
from .unet import *
from .pix2pix import *
from .pix2pixhd import *
from .autogan import AutoGAN, autogan_32, autogan_64, autogan_128
from .snres_discr import *
from .hourglass import Hourglass
from .attention import Attention56Bone, attention56
from .stylegan2 import StyleGAN2Generator, StyleGAN2Discriminator
from .efficient import EfficientNet
